#include <stdio.h>
#include <stdlib.h>
#include <sched.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include <sys/mman.h>
#include <sys/time.h>

#include "hack-gtod.h"

#define TDIFF(t0,t1) ((t1.tv_sec - t0.tv_sec)*1000000 + t1.tv_usec-t0.tv_usec)
#define NTRIALS 10

/*
 * Increment the argument, in order to step 32 times each octave:
 * this means we step 1k from 32k to 64k, but we step 512 from 16k to 32k.
 * There is no interest in optimizing this
 */
int incr(int n)
{
    int step;
    if (n < 64) return n + 1;
    /* find the most significant bit; */
    for (step = 1; step <= n; step <<= 1)
	;
    step = step >> 6;
    return n + step;
}

int main(int argc, char **argv)
{
    struct sched_param param;
    struct timeval tv1,tv2;
    unsigned char *b1, *b2;
    int maxmegs, i, j;
    long l, min, max, gtod, mc;
    long long tot;

    if (argc != 2)  exit(1);
    maxmegs = atoi(argv[1]);
    if (!maxmegs) exit(2);

    if (mlockall(MCL_CURRENT | MCL_FUTURE)) {
        fprintf(stderr, "%s: mlockall: %s\n", argv[0], strerror(errno));
        /* continue anyway */
    }
    if (!getenv("BENCH_NOTRT")) {
	param.sched_priority = 1;
	if (sched_setscheduler(0, SCHED_RR, &param)<0) {
	    fprintf(stderr,"%s: setscheduler: %s\n", argv[0], strerror(errno));
	    /* continue anyway */
	}
    }

    b1 = malloc(maxmegs << 20);
    b2 = malloc(maxmegs << 20);

    gettimeofday(&tv1, NULL);
    gettimeofday(&tv1, NULL);
    gettimeofday(&tv2, NULL);
    gtod = TDIFF(tv1,tv2);
    fprintf(stderr, "gettimeofday takes %li usecs\n", gtod);

    memcpy(b1, b2, 0);
    gettimeofday(&tv1, NULL);
    memcpy(b1, b2, 0);
    gettimeofday(&tv2, NULL);
    mc = TDIFF(tv1,tv2) - gtod;
    if (mc < 0) mc = 0;
    fprintf(stderr, "memcpy(0)    takes %li usecs\n", mc);

    /* ok, now memcpy stuff, NTRIALS times each */
    for (i = 1; i <= maxmegs<<20; i = incr(i)) {
	min = LONG_MAX, max = 0, tot = 0;
	for (j = 0; j < NTRIALS; j++) {
	    gettimeofday(&tv1, NULL);
	    memcpy(b1, b2, i);
	    gettimeofday(&tv2, NULL);
	    l = TDIFF(tv1, tv2) - gtod - mc;
	    if (l < 0 ) l = 0;
	    if (l < min) min = l;
	    if (l > max) max = l; /* max currently not used */
	    tot = tot + l; /* tot (avg) currently not used */
	}
        printf("%9i    %9li %9f\n", i, min, (double)min/(double)i);
    }
    return 0;
}
	    










